/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									     *
 * Library:	libsubdiv2.5D						     *
 *									     *
 * Module:	subdiv2.5D_txitxa.c					     *
 *									     *
 * Description:	The Real Thing (main functions for subdivision surfaces)     *
 *									     *
 * Date:	20000314						     *
 *									     *
 * Author:	Francisco MORN BURGOS   (http://www.gti.ssr.upm.es./~fmb/)  *
 *		Grupo Tratamiento Imgenes, Univ. Politcnica Madrid, Spain  *
 *									     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include <string.h>

#include "subdiv2.5D_priv.h"



int	sdP1[3] = {1, 2, 0},	/* sdP1[i] = (i+1) % 3 for i = {0,1,2} */
	sdM1[3] = {2, 0, 1};	/* sdM1[i] = (i-1) % 3 for i = {0,1,2} */

sdVector sdNullVector = {0.0F, 0.0F, 0.0F};



void sdSetEpsilon(sdMeshPtr m, float epsilon)
{
  if (!m)
    return;

  if (m->progressive)
    sdiError("sdSetEpsilon: cannot have both progressive and adaptive modes",
	     -1);
  else if (0 >= epsilon || epsilon >= 1)
    sdiError("sdSetEpsilon: epsilon must lie strictly between 0 and 1", -1);
  else
    m->epsilon = epsilon;
} /* sdSetEpsilon */


void sdSetProgressiveMode(sdMeshPtr m, sdBoolean progressive)
{
  if (!m)
    return;

  if (m->epsilon > 0)
    sdiError("sdSetProgressiveMode: cannot have both progressive and adaptive "
	     "modes", -1);
  else
    m->progressive = progressive;
} /* sdSetProgressiveMode */


#define SON_OF_T_WHOSE_ADT_IS_T(ta, t)				\
  ta->son[ta->son[0]->adt[2]==t || ta->son[0]->adt[0]==t? 0 :	\
	  ta->son[1]->adt[0]==t || ta->son[1]->adt[1]==t? 1 :	\
	  ta->son[2]->adt[1]==t || ta->son[2]->adt[2]==t? 2 :	\
	  sdiError("SON_OF_T_WHOSE_ADT_IS_T: yuyu..", -8)]

#define SUBST_ADT_T(t, to, tn)	\
  t->adt[t->adt[0]==to? 0 :	\
	 t->adt[1]==to? 1 :	\
	 t->adt[2]==to? 2 :	\
	 sdiError("SUBST_ADT_T: yuyu..", -8)] = tn

void sdMergeTriangle(sdMeshPtr m, sdTrianglePtr t)
{
  sdTrianglePtr tm, ta, tas, ts, tsa;
  int i, k;
  sdDetailTree dt;

  if (!m || !t || !t->mom)
    return;

  /* 1. make sure that the generational gap is never deeper than one	*/
  /*    and fix adjacency information					*/
  tm = t->mom;
  for (k=0; k<4; k++)
    if (tm->son[k]->subdiv)
      sdMergeTriangle(m, tm->son[k]->son[3]);

  for (i=0; i<3; i++)
    if ((ta = tm->adt[i]) && ta->subdiv) {
      tas = SON_OF_T_WHOSE_ADT_IS_T(ta, tm->son[i]);
      if (tas->subdiv)
	sdMergeTriangle(m, tas->son[3]);
      SUBST_ADT_T(tas, tm->son[i], tm);
      tas = SON_OF_T_WHOSE_ADT_IS_T(ta, tm->son[sdP1[i]]);
      if (tas->subdiv)
	sdMergeTriangle(m, tas->son[3]);
      SUBST_ADT_T(tas, tm->son[sdP1[i]], tm);
    }

  /* 2. delete [eventually detail trees,] points and triangles		*/
  if (m->progressive && !tm->inv) {
    dt = tm->dt;
    for (k=0; k<4; k++)
      sdlDeleteLeafDetailTree(&dt->son[k]);
  }
  for (k=0; k<3; k++) {
    ts = tm->son[k];
    if (!(tsa = ts->adt[k]) || tsa->depth < ts->depth) {
      sdlDeletePointFromVertexList(m, &ts->vtx[sdP1[k]]);
    }
    sdlDeleteLeafTriangle(m, &ts);
  }
  sdlDeleteLeafTriangle(m, &tm->son[3]);
  DEL(tm->son);
  tm->subdiv = FALSE; m->NleafTriangles++;
  m->refresh = TRUE;
} /* sdMergeTriangle */


void sdMergeAllLeavesOfTriangle(sdMeshPtr m, sdTrianglePtr t)
{
  int k;

  if (!m || !t || !t->subdiv)
    return;

  if (!t->son[3]->subdiv)
    sdMergeTriangle(m, t->son[3]);
  else
    for (k=0; k<4; k++)
      sdMergeAllLeavesOfTriangle(m, t->son[k]);
} /* sdMergeAllLeavesOfTriangle */


void sdMergeAllLeafTrianglesOnce(sdMeshPtr m)
{
  int i;

  if (!m)
    return;

  if (m->verbose)
    fprintf(stderr, "Merging     %6d leaf triangles...  ", m->NleafTriangles);
  for (i=0; i<m->NmotherTriangles; i++)
    sdMergeAllLeavesOfTriangle(m, m->motherTriangle[i]);
  if (--m->depth < 0)
    m->depth = 0;
  if (m->verbose)
    fprintf(stderr, "Done.\n");
} /* sdMergeAllLeafTrianglesOnce */


typedef enum {DIRECT, GENERATIONAL_GAP, BUTTERFLY_NBRS} sdtKindOfCall;

static void sdtSubdivTriangle(sdMeshPtr m, sdTrianglePtr t, sdtKindOfCall call)
{
  int i, j, d;
  sdPointPtr p[3], mp[3];
  sdTrianglePtr ta[3], taa, *ts;
  sdBoolean ta_subdiv[3], fc2BM, mp_new[3];
  sdVector v[3], nv[3], v_inv[3], detail;
  sdDetailTree dt, *ds;
  char ids[23], *idsaux;
  sdTriangleLst tl;
  float k;
#ifdef DIME
  float v1[3], v2[3];
#endif

  if (!m || !t || t->subdiv)
    return;

  /* 0. try to avoid working...						*/
  if (call == DIRECT && !m->progressive &&
      (t->K < 555*EPSILON ||
       m->epsilon > 0 && sdmCoplanarNeighbours(t, 1.0F - m->epsilon)))
    return;

  t->subdiv = TRUE; m->NleafTriangles--;
  m->refresh = TRUE;

  /* 1. make sure that there's never a generational gap deeper than one	*/
  /*    generation [and, eventually, that all neighbours needed for the	*/
  /*	butterfly scheme exist before trying to build the new vertices;	*/
  /*	keep on a list triangles built only to satisfy butterfly scheme	*/
  /*	needs, so that they can be deleted later in step 7]		*/
  for (i=0; i<3; i++) {
    p[i] = t->vtx[i];
    if (!(ta[i] = t->adt[i]))
      ta_subdiv[i] = FALSE;
    else {
      if (ta_subdiv[i] = (ta[i]->depth < t->depth)) {
	if (call != BUTTERFLY_NBRS) 
	  sdtSubdivTriangle(m, ta[i], GENERATIONAL_GAP);
	else {
          sdtSubdivTriangle(m, ta[i], BUTTERFLY_NBRS);
	  sdlAddTriangle2TriangleList(ta[i]->son[3], &m->tmpTriangleList);
	}
      }
      if (m->subdivScheme == BUTTERFLY) {
	j = IDX_VTX_IN_T(p[i], t->adt[i]);
	if ((taa = t->adt[i]->adt[j]) && taa->depth < t->depth) {
	  sdtSubdivTriangle(m, taa, BUTTERFLY_NBRS);
	  sdlAddTriangle2TriangleList(taa->son[3], &m->tmpTriangleList);
	}
	if ((taa = t->adt[i]->adt[sdP1[j]]) && taa->depth < t->depth) {
	  sdtSubdivTriangle(m, taa, BUTTERFLY_NBRS);
	  sdlAddTriangle2TriangleList(taa->son[3], &m->tmpTriangleList);
	}
      }
    }
  }

  /* 2. build new vertices (with normals) and add them to the points	*/
  /*	list, or simply take them from already subdivided neighbours	*/
  /*	of this triangle, according to the following layout:		*/
  /*									*/
  /*				  p2					*/
  /*				  / \					*/
  /*				 /   \					*/
  /*				/     \					*/
  /*			       /       \				*/
  /*			     mp2-------mp1				*/
  /*			     / \       / \				*/
  /*			    /	\     /	  \				*/
  /*			   /	 \   /	   \				*/
  /*			  /	  \ /	    \				*/
  /*			p0--------mp0--------p1				*/
  /*									*/
  for (i=0; i<3; i++) {
    v[i] = p[i]->pos;
    nv[i] = p[i]->normal;
  }
  for (i=0, fc2BM=TRUE; i<3; i++) {
    if (ta[i] && !ta_subdiv[i] && ta[i]->subdiv) {
      j = IDX_VTX_IN_T(p[sdP1[i]], ta[i]);
      mp[i] = ta[i]->son[j]->vtx[sdP1[j]];
      mp_new[i] = FALSE;
    } else {
      switch (m->subdivScheme) {
	case BUTTERFLY:
	  mp[i] = sdlAddPoint2VertexList(m, sdmButterflyMidpoint(t, i, fc2BM));
          /* midpoint normal = simple average of endpoints normals	*/
	  /* ToDo: take into account surrounding triangles/vertices	*/
	  mp[i]->normal = *sdmNormalise(sdmMidpoint(nv+i, nv+sdP1[i]));
	  fc2BM = FALSE;
	  break;
        case MIDPOINT:
	  mp[i] = sdlAddPoint2VertexList(m, sdmMidpoint(v+i, v+sdP1[i]));
          /* midpoint normal = simple average of endpoints normals	*/
	  /* ToDo: take into account surrounding triangles/vertices	*/
	  mp[i]->normal = *sdmNormalise(sdmMidpoint(nv+i, nv+sdP1[i]));
	  break;
	case SPHERE:
	  nv[i] = *sdmNormalise(sdmMidpoint(v+i, v+sdP1[i]));
	  mp[i] = sdlAddPoint2VertexList(m, nv+i);
	  mp[i]->normal = mp[i]->pos;
	  break;
      }
      mp_new[i] = TRUE;
    }
  }

  /* 3. build new triangles according to the following layout, in which	*/
  /*	the dots mark the vertex nr. 0 of the new triangles:		*/
  /*									*/
  /*				  p2					*/
  /*				  / \					*/
  /*				 /   \					*/
  /*				/ ts2 \					*/
  /*			       /      \				*/
  /*			     mp2-------mp1				*/
  /*			     / \      / \				*/
  /*			    /	\ ts3 /	  \				*/
  /*			   / ts0 \   / ts1 \				*/
  /*			  /	  \ /	    \				*/
  /*			p0--------mp0--------p1				*/
  /*									*/
  if (!(ts = t->son = NEW(4, sdTrianglePtr)))
    sdiError("subdivTriangle: not enough memory for a new set of triangles",
	     -13);
  if ((d = t->depth + 1) > m->depth)
    m->depth = d;
  for (i=0; i<4; i++)
    ts[i] = sdlNewLeafTriangle(m, d);
  ts[0]->vtx[0] = p[0];  ts[0]->vtx[1] = mp[0]; ts[0]->vtx[2] = mp[2];
  ts[1]->vtx[0] = mp[0]; ts[1]->vtx[1] = p[1];  ts[1]->vtx[2] = mp[1];
  ts[2]->vtx[0] = mp[2]; ts[2]->vtx[1] = mp[1]; ts[2]->vtx[2] = p[2];
  ts[3]->vtx[0] = mp[1]; ts[3]->vtx[1] = mp[2]; ts[3]->vtx[2] = mp[0];
  mp[0]->t = mp[1]->t = mp[2]->t = ts[3];
  ts[0]->inv = ts[1]->inv = ts[2]->inv = t->inv; ts[3]->inv = !t->inv;

  /* 4. build hierarchy and adjacency information for new triangles	*/
  for (i=0; i<4; i++)
    ts[i]->mom = t;
  for (i=0; i<3; i++) {
    if (ta_subdiv[i]) {
      ts[i]->adt[i] = ts[sdP1[i]]->adt[i] = SON_OF_T_WHOSE_ADT_IS_T(ta[i], t);
    } else if (ta[i] && ta[i]->subdiv) {
      j = IDX_VTX_IN_T(p[sdP1[i]], ta[i]);
      ts[i]->adt[i] = ta[i]->son[sdP1[j]]; ta[i]->son[sdP1[j]]->adt[j] = ts[i];
      ts[sdP1[i]]->adt[i] = ta[i]->son[j]; ta[i]->son[j]->adt[j] = ts[sdP1[i]];
    } else {
      ts[i]->adt[i] = ts[sdP1[i]]->adt[i] = ta[i];
    }
    ts[i]->adt[sdP1[i]] = ts[3];
    ts[3]->adt[i] = ts[sdM1[i]];
  }

  /* 5. eventually build and plant detail trees according to the	*/
  /*	following layout and rules:					*/
  /*									*/
  /*				   ^	     				*/
  /*				  / \	     				*/
  /*				 /   \	      			*/
  /*				/ ts2 \	  ta1s3			*/
  /*			       /       \ 	 			*/
  /*			     ds2-------ds3    			*/
  /*			     / \       / \				*/
  /*			    /	\     /	  \				*/
  /*			   / ts0 \   / ts1 \				*/
  /*			  /	  \ /	    \				*/
  /*		     dt,ds0-------ds1--------				*/
  /*									*/
  /*	- if "t" is an inverted triangle, no new trees are needed, but	*/
  /*	  its normal son, "ts3", is one of the "ta1s3" described below:	*/
  /*	  attach it to the corresponding "ta1->ds3"			*/
  /*	- if "t" is a normal triangle, its detail tree "dt", must have	*/
  /*	  4 sons, of which:						*/
  /*	  + "ds{0,1,2}" must point respectively to "ts{0,1,2}->v0" and	*/
  /*	    "ts{0,1,2}" must point respectively to "ds{0,1,2}"		*/
  /*	  + "ds3" must point to "mp1" and:				*/
  /*	    * if "t" is a border triangle inside this "motherTriangle"	*/
  /*	      ("dt->id" has no "0's"), "ds3" will be sterile because no	*/
  /*	      triangle will ever point to it				*/
  /*	    * otherwise, a triangle "ta1s3", that'll be built whenever	*/
  /*	      "ta1" gets subdivided, will be attached to it; if "ta1s3"	*/
  /*	      gets subdivided itself later on, "ds3" will have sons	*/
  if (m->progressive) {
    if (t->inv)
      ts[3]->dt = ta[1]->dt->son[3];
    else {
      dt = t->dt;
      if (!(ds = dt->son = NEW(4, sdDetailTree)))
	sdiError("subdivTriangle: not enough memory for a new set of detail "
		 "trees", -13);
      idsaux = strcpy(ids, dt->id) + strlen(dt->id);
      *(idsaux+1) = '\0';
      for (i=0, *idsaux='0'; i<4; i++, (*idsaux)++) {
	ds[i] = sdlNewLeafDetailTree(ids);
	if (i<3) {
	  ds[i]->pt = ts[i]->vtx[0];
	  ts[i]->dt = ds[i];
	} 
      }
      ds[3]->pt = mp[1];
      if (!m->fDetailsIn)
	/* calculate details and keep track of "maxDetail{N,T}" */
	for (i=0; i<3; i++)
	  if (mp_new[i]) {
	    sdmBuildFrameForTriangleEdge(t, i, v);
	    if (m->targetMesh)
	      detail = *sdmCalculateDetail(m, mp[i], v);
	    else {
#ifdef DIME
	      if (m->useDiMePlugin) {
		v1[0]=mp[i]->pos.x; v1[1]=mp[i]->pos.y; v1[2]=mp[i]->pos.z;
		dimeCalculateDetail(v1, v2);
		detail.x = v2[0]; detail.y = v2[1]; detail.z = v2[2];
	      } else {	/* no target mesh => use unit sphere */
#endif
		detail = *sdmNormalise(&mp[i]->pos);
		DEC_VEC(detail, mp[i]->pos);
#ifdef DIME
	      }
#endif
	    }
	    sdmMatrixTrans(v, v_inv);	/* v orthonormal => inv = trans */
	    mp[i]->delta = *sdmMatrixVecProd(v_inv, &detail);
	    if ((k = ABS(mp[i]->delta.x)) > m->maxDetailN) m->maxDetailN = k;
	    if ((k = ABS(mp[i]->delta.y)) > m->maxDetailT) m->maxDetailT = k;
	    if ((k = ABS(mp[i]->delta.z)) > m->maxDetailT) m->maxDetailT = k;
	  }  
    }
  }

  /* 6. calculate normals and areas of new triangles -- this has to be	*/
  /*	done, of course, AFTER adding details to the new vertices, so	*/
  /*	it will have to wait until later if this is a progressive mesh	*/
  if (!m->progressive)
    for (i=0; i<4; i++)
      sdmFillTriangleNormalAndK(ts[i], TRUE);

  /* [7. eventually flush the temporary triangles list]			*/
  if (call == DIRECT && m->tmpTriangleList) {
    for (tl=m->tmpTriangleList; tl; tl=tl->next)
      sdMergeTriangle(m, tl->t);
    sdlDestroyTriangleList(&m->tmpTriangleList);
  }
} /* sdtSubdivTriangle */


void sdSubdivTriangle(sdMeshPtr m, sdTrianglePtr t)
{
  sdtSubdivTriangle(m, t, DIRECT);
} /* sdSubdivTriangle */


void sdSubdivAllLeavesOfTriangle(sdMeshPtr m, sdTrianglePtr t)
{
  int k;

  if (!m || !t)
    return;

  if (!t->subdiv)
    sdtSubdivTriangle(m, t, DIRECT);
  else
    if (!m->progressive || !t->inv)
      for (k=0; k<4; k++)
	sdSubdivAllLeavesOfTriangle(m, t->son[k]);
    else
      for (k=3; k>=0; k--)
	sdSubdivAllLeavesOfTriangle(m, t->son[k]);
} /* sdSubdivAllLeavesOfTriangle */


static void sdtReposnMidptsOfTriangleAtDepth(sdMeshPtr m, sdTrianglePtr t,
					     int depth, int nbits,
					     sdBoolean recalcPosns,
					     sdBoolean recalcDetails,
					     sdBoolean addDetails)
{
  int i;
  sdVectorPtr vtx[3];
  sdTrianglePtr ta;
  sdBoolean fc2BM;
  static int k[3] = {1, 3, 2};
  sdPointPtr mpt;
  sdVector pickedVertex_pos, v[3], v_inv[3], detail;
#ifdef DIME
  float v1[3], v2[3];
#endif

  if (!m || !m->progressive || !t || !t->subdiv)
    return;

  if (t->depth < depth)
    for (i=0; i<4; i++)
      sdtReposnMidptsOfTriangleAtDepth(m, t->son[i], depth, nbits, recalcPosns,
				       recalcDetails, addDetails);
  else {
    for (i=0; i<3; i++)
      vtx[i] = &t->vtx[i]->pos;
    for (i=0, fc2BM=TRUE; i<3; i++)
      if (!(ta = t->adt[i]) || t->n < ta->n) {
	if ((mpt = t->son[k[i]]->vtx[0]) == m->pickedVertex)
	  pickedVertex_pos = mpt->pos;
	if (recalcPosns)
	  switch (m->subdivScheme) {
	  case BUTTERFLY:
	    mpt->pos = *sdmButterflyMidpoint(t, i, fc2BM);
	    fc2BM = FALSE;
	    break;
	  case MIDPOINT:
	    mpt->pos = *sdmMidpoint(vtx[i], vtx[sdP1[i]]);
	    break;
	  case SPHERE:
	    mpt->pos = *sdmNormalise(sdmMidpoint(vtx[i], vtx[sdP1[i]]));
	    break;
	  }
	if (addDetails) {
	  sdmBuildFrameForTriangleEdge(t, i, v);
	  if (!m->fDetailsIn) {
	    if (mpt == m->pickedVertex) {
	      detail = pickedVertex_pos;
	      DEC_VEC(detail, mpt->pos);
 	      sdmMatrixTrans(v, v_inv);  /* v orthonormal => inv = trans */
	      mpt->delta = *sdmMatrixVecProd(v_inv, &detail);
	    }
	    if (recalcDetails) {
	      if (m->targetMesh)
		detail = *sdmCalculateDetail(m, mpt, v);
	      else {
#ifdef DIME
		if (m->useDiMePlugin) {
		  v1[0] = mpt->pos.x; v1[1] = mpt->pos.y; v1[2] = mpt->pos.z;
		  dimeCalculateDetail(v1, v2);
		  detail.x = v2[0]; detail.y = v2[1]; detail.z = v2[2];
		} else {	/* no target mesh => use unit sphere */
#endif
		  detail = *sdmNormalise(&mpt->pos);
		  DEC_VEC(detail, mpt->pos);
#ifdef DIME
		}
#endif
	      }
 	      sdmMatrixTrans(v, v_inv);  /* v orthonormal => inv = trans */
	      mpt->delta = *sdmMatrixVecProd(v_inv, &detail);
	    }
	    sdmEncodeDetail(m, nbits, &mpt->delta, &mpt->codedelta);
	  }
	  sdmDecodeDetail(m, nbits, mpt->codedelta, &mpt->delta);
	  detail = *sdmMatrixVecProd(v, &mpt->delta);
	  INC_VEC(mpt->pos, detail);
	}
      }
    m->refresh = TRUE;
  }
} /* sdtReposnMidptsOfTriangleAtDepth */


static void sdtRecalculateNormalsOfTriangleAtDepth(sdTrianglePtr t, int depth)
{
  int i;

  if (!t)
    return;

  if (t->depth < depth && t->son)
    for (i=0; i<4; i++)
      sdtRecalculateNormalsOfTriangleAtDepth(t->son[i], depth);
  else
    sdmFillTriangleNormalAndK(t, TRUE);
} /* sdtRecalculateNormalsOfTriangleAtDepth */


void sdRecalculateAllVertexPositions(sdMeshPtr m, int levelStartReposn,
				     int levelStopDetails, int nbits,
				     sdBoolean recalcDetails)
{
  int d, i;

  if (!m)
    return;

  for (d=levelStartReposn; d<m->depth; d++) {
    for (i=0; i<m->NmotherTriangles; i++)
      sdtReposnMidptsOfTriangleAtDepth(m, m->motherTriangle[i], d, nbits, TRUE,
				       recalcDetails, d < levelStopDetails);
    for (i=0; i<m->NmotherTriangles; i++)
      sdtRecalculateNormalsOfTriangleAtDepth(m->motherTriangle[i], d+1);
  }
} /* sdRecalculateAllVertexPositions */


void sdSubdivAllLeafTrianglesOnce(sdMeshPtr m)
{
  int i;
  sdBoolean pct_ind = FALSE;
  float pct, pct_step, MN, MT;

  if (!m)
    return;

  if (m->verbose) {
    fprintf(stderr, "Subdividing %6d leaf triangles...  ", m->NleafTriangles);
    if (pct_ind = (m->NleafTriangles > 1000)) {
      pct = 0;
      pct_step = 100.0F / m->NmotherTriangles;
    }
  }
  if (m->progressive && !m->fDetailsIn) {
    MN = m->maxDetailN;  MT = m->maxDetailT;
  }
  for (i=0; i<m->NmotherTriangles; i++) {
    if (pct_ind)
      fprintf(stderr, "%2d%% done...", (int) pct);
    sdSubdivAllLeavesOfTriangle(m, m->motherTriangle[i]);
    if (pct_ind) {
      fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b");
      pct += pct_step;
    }
  }

  if (m->progressive) {
    if (!m->fDetailsIn &&
	(m->depth == 1 || m->maxDetailN > MN || m->maxDetailT > MT)) {
      if (m->verbose > 1) {
	BEEP;
	fprintf(stderr, "Done.      \n"
			"  New (maxDetailN, maxDetailT) = (%+E; %+E)\n"
			"  => repositioning all vertices...  ",
		m->maxDetailN, m->maxDetailT);
      }
      sdRecalculateAllVertexPositions(m, 0, m->depth, NBPDN, TRUE);
    } else {
      if (m->verbose > 1)
	fprintf(stderr, "Done.      \n"
			"  Repositioning new vertices and calculating new "
			"triangle normals...  ");
      for (i=0; i<m->NmotherTriangles; i++)
	sdtReposnMidptsOfTriangleAtDepth(m, m->motherTriangle[i], m->depth-1,
					 NBPDN, FALSE, FALSE, !m->fDetailsIn);
      for (i=0; i<m->NmotherTriangles; i++)
	sdtRecalculateNormalsOfTriangleAtDepth(m->motherTriangle[i], m->depth);
    }
  }
  if (m->verbose)
    fprintf(stderr, "Done.      \n");
} /* sdSubdivAllLeafTrianglesOnce */


void sdSubdivAllLeafTrianglesUntilSmooth(sdMeshPtr m)
{
  int nlt;

  if (!m)
    return;

  if (m->progressive)
    sdiError("sdSubdivAllLeafTrianglesUntilSmooth: won't work in progressive "
	     "mode", -1);
  do {
    nlt = m->NleafTriangles;
    sdSubdivAllLeafTrianglesOnce(m);
  } while (m->NleafTriangles > nlt);
} /* sdSubdivAllLeafTrianglesUntilSmooth */


/* To understand the following function ("sdSPIHTcodec"), read:		*/
/* 1. Said & Pearlman: "A New, Fast and Efficient Image Codec Based on	*/
/*    Set Partitioning in Hierarchical Trees", IEEE Trans. on Circuits	*/
/*    and Systems for Video Technology, 6(3):243-250, 1996;		*/
/* 2. Kolarov & Lynch: "Compression of Scalar Functions Defined on	*/
/*    2-manifolds", Interval Research Corp. internal report, 1996.	*/

#ifdef BIT_SHUFFLING
#  define MSKINIT	MASK_SN >> 3
#  define NINIT		NBPD - 4
#else
#  define MSKINIT	MASK_SN >> 1
#  define NINIT		NBPD - 2
#endif

#define IO_NEXT_BIT(b, cw, mask)				\
  if (m->fDetailsIn) {						\
    if (b = sdiReadNextBitFromDetailsFile(m)) cw |= mask;	\
  } else {							\
    sdiWriteNextBitToDetailsFile(m, b = (cw & mask));		\
  }

void sdSPIHTcodec(sdMeshPtr m, sdBoolean animate)
{
  int n, i, nbits_sort, nbits_reft, Nbits;
  sdDTreeList lit, lit_, lid, lid_, lsd, lsd_, dtl, dtl_next;
  CODEWORD mask;
  sdBoolean b, little_children;
  sdDetailTree ds[4];

  if (!m || !m->progressive)
    return;

  /* 0.	Initialization:							*/
  /*	- lit (list of insignificant trees) must contain all mother	*/
  /*	      detail trees, which are inserted as type A (+) entries	*/
  /*	- lid (list of insignificant details) and			*/
  /*	- lsd (list of significant details) must be empty		*/
  lit = lit_ = lid = lid_ = lsd = lsd_ = NULL;
  for (i=0; i<m->NmotherTriangles; i++)
    if (m->motherDetailTree[i]->son)
      sdlAddDetailTree2DTreeList(m->motherDetailTree[i], +NINIT, &lit, &lit_);
  Nbits = nbits_sort = nbits_reft = 0;

  for (mask=MSKINIT, n=NINIT; mask; mask>>=1, n--) {

    /* 1. Sorting -- the sign bits that "IO_NEXT_BIT(b, ..., MASK_SN)"	*/
    /*	  writes/reads are accounted for as refinement bits		*/
    for (dtl=lid; dtl; dtl=dtl_next) {
      dtl_next = dtl->next;
      IO_NEXT_BIT(b, dtl->dt->pt->codedelta, mask);  nbits_sort++;
      if (b) {
	IO_NEXT_BIT(b, dtl->dt->pt->codedelta, MASK_SN);  nbits_reft++;
	sdlAddDetailTree2DTreeList(dtl->dt, n, &lsd, &lsd_);
	sdlDeleteDTreeListFromDTreeList(&dtl, &lid, &lid_);
      }
    }
    for (dtl=lit; dtl; dtl=dtl_next) {
      dtl_next = dtl->next;
      if (dtl->n > 0) {	/* type A (+) entry */
	IO_NEXT_BIT(b, dtl->dt->D, mask);  nbits_sort++;
	if (b && dtl->dt->son) {
	  for (i=1, little_children=FALSE; i<4; i++) {
	    ds[i] = dtl->dt->son[i];
	    /* there's no need to put "ds[0]" in either "lsd" or "lid",	*/
	    /* since it points to the same vertex as "dtl->dt"		*/
	    if (ds[i]->son) little_children = TRUE;
	    IO_NEXT_BIT(b, ds[i]->pt->codedelta, mask);  nbits_sort++;
	    if (b) {
	      IO_NEXT_BIT(b, ds[i]->pt->codedelta, MASK_SN);  nbits_reft++;
	      sdlAddDetailTree2DTreeList(ds[i], n, &lsd, &lsd_);
	    } else
	      sdlAddDetailTree2DTreeList(ds[i], 0, &lid, &lid_);
	  }
	  if (little_children)
	    sdlAddDetailTree2DTreeList(dtl->dt, -n, &lit, &lit_);
	  sdlDeleteDTreeListFromDTreeList(&dtl, &lit, &lit_);
	}
      } else {		/* type B (-) entry */
	IO_NEXT_BIT(b, dtl->dt->L, mask);  nbits_sort++;
	if (b) {
	  for (i=0; i<4; i++)
	    sdlAddDetailTree2DTreeList(dtl->dt->son[i], +n, &lit, &lit_);
	  sdlDeleteDTreeListFromDTreeList(&dtl, &lit, &lit_);
	}
      }
    }

    /* 2. Refinement */
    for (dtl=lsd; dtl; dtl=dtl->next)
      if (dtl->n > n) {
	IO_NEXT_BIT(b, dtl->dt->pt->codedelta, mask);  nbits_reft++;
      }

    /* 3. [Eventual mesh update animation and "accounting"] */
    if (animate) {
      sdRecalculateAllVertexPositions(m, 0, m->depth, NBPDN, FALSE);
      sdgDisplayMainMesh();
    }
    if (m->verbose) {
      Nbits += nbits_sort + nbits_reft;
      if (m->verbose == 1) {
	if (n == NINIT)
	  fprintf(stderr, "\n  (bitplane, #bits sort, #bits reft) = ");
	else
	  fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
	fprintf(stderr, "(%2d, %5d, %5d)", n, nbits_sort, nbits_reft);
	if (n == 0)
	  fprintf(stderr, " => Total: %d bits\n", Nbits);
      } else {
	if (n == NINIT)
	  fprintf(stderr, "\n  {(bitplane, #bits sort, #bits reft)} = {\n");
	else
	  fprintf(stderr, "; ");
	fprintf(stderr, "(%2d, %5d, %5d)", n, nbits_sort, nbits_reft);
	if (n == 0)
	  fprintf(stderr, "\n} => Total: %d bits\n", Nbits);
      }
      nbits_sort = nbits_reft = 0;
    }

  }

  sdlDestroyDTreeList(&lit);
  sdlDestroyDTreeList(&lid);
  sdlDestroyDTreeList(&lsd);
} /* sdSPIHTcodec */

